/********************************************************************

  >>>>> USE SEPARATION TRANSFORMER FOR TESTING <<<<<<

- Library for soft serial: https://github.com/plerup/espsoftwareserial
- Used ESP8266 version is Wemos D1 mini

- Power meter version is meter with V9211F chip from Vango

V01   20170606
********************************************************************/
#include <ArduinoOTA.h>
#include <PubSubClient.h> 
#include <SoftwareSerial.h>

#include "SSIDinfo.h"
 
byte rx;
byte state;
byte bytecount, ix;
byte data[20];
byte checksum;
unsigned long prevmillis, lastSend;
boolean datafound;
float volt, avgVolt;
float watt, avgWatt;
float freq, avgFreq;
int countFreq;
int numReadings, freqReadings;
float timeFreq;
unsigned long prevMicros;
int testcounter;

const int sync_interval = 600;           // check for synchronisation
const int send_interval = 60*1000;      // 1 minute interval for mqtt messages

WiFiClient wifiClient;
void callback(char* topic, byte* payload, unsigned int length)
  {delay(1);}                           // do nothing
PubSubClient client(MQTT_SERVER, 1883, callback, wifiClient); 

SoftwareSerial softRx(D1, SW_SERIAL_UNUSED_PIN, false, 256);    // TX on V9211F pcb

/********************************************************************/
void setup()
  {
  Serial.begin(115200);
  softRx.begin(4800);
  delay(100);
  Serial.println("\nSoftware serial started");

  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);                   // start wifi subsystem
  WiFi.config(ip, gateway, subnet);             // attempt to connect to the WIFI network
  reconnect();                                  // and then connect to the MQTT server
  delay(100);
  
  ArduinoOTA.setHostname(hostName);
  ArduinoOTA.setPassword((const char *) upload_w8w); 
  ArduinoOTA.onStart([]() { delay(50); });
  ArduinoOTA.onEnd([]() { delay(50); });
  ArduinoOTA.onError([](ota_error_t error) { ESP.restart(); });
  ArduinoOTA.begin();

  attachInterrupt(D2, freqinthandler, RISING);

  state = 0;
  datafound = false;
  prevmillis = lastSend = millis();
  }

/********************************************************************/
void loop()
  {
  ArduinoOTA.handle();
  if (client.connected()) {client.loop();}      // maintain MQTT connection
  else if (WiFi.status() == 3) {reconnect();}
  
  switch (state)
    {
    case 0: {                                   // waiting for begin info power meter
            while (softRx.available() > 0)      // check for data
              {
              softRx.flush();
              datafound = true;
              prevmillis = millis();
              }
                                                // if so wait for gap of 600 msec
            if (((millis() - prevmillis) > sync_interval) && datafound)
              {
              softRx.flush();
              bytecount = 0;
              state = 1;
              }
             } break;

    case 1: {                                   // 8 bytes SysCtrl skipped
            while (softRx.available() > 0)
              {
              rx = softRx.read();
              if (bytecount++ >= 7)
                {
                softRx.flush();
                bytecount = 0;
                state = 2;
                }
              }
            } break;

    case 2: {                                   // 20 bytes, including power and voltage data
            while (softRx.available() > 0)
              {
              rx = softRx.read();
              data[bytecount] = rx;
              if (bytecount++ >= 19)
                {
                softRx.flush();
                prevmillis = millis();
                bytecount = 0;
                state = 3;
                }
              }
            } break;
          
    case 3: {                                   // process and send data
            if (checksumOK())
              {
              volt = 0;                         // 3rd word data is volt
              volt = ((data[14] << 24) + (data[13] << 16) + (data[12] << 8) + data[11]);
              volt = volt / 4154255;

              watt = 0;                         // 1rst word data is watt
              watt = ((data[6] << 24) + (data[5] << 16) + (data[4] << 8) + data[3]);
              watt = watt / 150775;
              if (watt < 0) {watt = 0;}

              if ((volt > 200) && (volt < 260) && (watt >= 0) && (watt < 3880)) // ignore spurious readings
                {
                numReadings++;                 // increment for average value calculation
                avgVolt = (volt + (avgVolt * (numReadings-1))) / numReadings;
                avgWatt = (watt + (avgWatt * (numReadings-1))) / numReadings; 
                }
                
              timeFreq = timeFreq / countFreq;
              countFreq = 0;
              if ((timeFreq > 15000) && (timeFreq <25000))
                {
                freqReadings++;                 // increment for average value calculation
                freq = 1000000 / timeFreq;
                timeFreq = 0;
                avgFreq = (freq + (avgFreq * (freqReadings-1))) / freqReadings; 
                }
              }
              else Serial.println("checksum not ok");

            if ((millis() - lastSend) > send_interval)
              {
              testcounter++;
              if (testcounter < 100) Serial.print('0');
              if (testcounter < 10) Serial.print('0');
              Serial.print(testcounter);
              Serial.print(": ");

              Serial.print(avgVolt, 1);
              Serial.print(" volt - ");
              Serial.print(avgWatt, 1);
              Serial.print(" watt - ");
              Serial.print(avgFreq, 1);
              Serial.print(" Hz");
              publishMqtt(); 
              numReadings = 0;
              freqReadings = 0;
              lastSend = millis(); 
              }
              
            prevmillis = millis();
            bytecount = 0;
            state = 4;
            } break;

    case 4: {                                 // if out of sync, sync here
            if (softRx.available() > 0)
              {
              softRx.flush();
              prevmillis = millis();
              }

            if (millis() - prevmillis > sync_interval)
              {
              state = 1;
              }
            } break;
          
    default:  break;
    }
  }

/********************************************************************/
boolean checksumOK()
  {
  checksum = 0;
  for (ix = 0; ix < 19; ix++) {checksum = checksum + data[ix];}
  checksum = ~checksum + 0x33;
  if (checksum == data[19]) {return true;}
    else {return false;}
  }

/*******************************************/
void publishMqtt()
  {
  long startmillis = millis();
  char payloadVal[10];

  String val = String(avgVolt, 1); 
  val.toCharArray(payloadVal, val.length() + 1); 
  client.publish("/pwrmtr2/volt", payloadVal);

  val = String(avgWatt, 1);
  val.toCharArray(payloadVal, val.length() + 1); 
  client.publish("/pwrmtr2/watt", payloadVal);

  val = String(avgFreq, 1);
  val.toCharArray(payloadVal, val.length() + 1); 
  client.publish("/pwrmtr2/freq", payloadVal);

  Serial.print("   ->  mqtt ");              
  Serial.println(millis() - startmillis);              
  }

/*********************************************/
// subroutines
void reconnect()                              //attempt to connect to the wifi if connection is lost
  {
  if(WiFi.status() != WL_CONNECTED)
    {
    Serial.print("\nConnecting to ");
    Serial.print(ssid);
    Serial.print(" ... ");
    while (WiFi.status() != WL_CONNECTED) { //loop while we wait for connection
      delay(100);
      }
    Serial.println("connected");  
    }

// be sure to be connected to WIFI before (re)connect to MQTT
  if (WiFi.status() == WL_CONNECTED)        // Loop until we're reconnected to the MQTT server
    {
    while (!client.connected())
      {
      Serial.print("Connecting to MQTT server ... ");
      String clientName;
      clientName += hostName;
      clientName += "-";
      uint8_t mac[6];
      WiFi.macAddress(mac);
      clientName += macToStr(mac);
      if (client.connect((char*) clientName.c_str()))
        {
        Serial.println("connected");
        Serial.print(hostName);
        Serial.print(" @ ");
        Serial.println(ip);
        }
      else {Serial.println("\tFailed."); abort();}
      }
    }
  }

/*********************************************/
//generate unique name from MAC addr
String macToStr(const uint8_t* mac)
  {
  String result;
  for (int i = 0; i < 6; ++i)
    {
    result += String(mac[i], 16);
    if (i < 5){result += ':';} 
    }
  return result;
  }

/*********************************************/
void freqinthandler()
  {
  countFreq++;
  timeFreq = timeFreq + micros() - prevMicros;
  prevMicros = micros();
  }

/********************************************************************/

